package com.persagy.apm.report.indicator.factory;

import com.google.gson.Gson;
import com.persagy.apm.report.common.constants.DataConstants;
import com.persagy.apm.report.common.utils.DataUtils;
import com.persagy.apm.report.common.utils.DateUtils;
import com.persagy.apm.report.dependencies.energyefficiencyweb.model.vo.EnergyEfficiencyDataVO;
import com.persagy.apm.report.dependencies.energyefficiencyweb.service.EnergyEfficiencyServiceImpl;
import com.persagy.apm.report.detail.enums.AttrValueTypeEnum;
import com.persagy.apm.report.detail.model.vo.AttrValueVO;
import com.persagy.apm.report.detail.model.vo.ItemVO;
import com.persagy.apm.report.indicator.codes.energyefficiency.AreaEnergyEfficiencyContentCodes;
import com.persagy.apm.report.indicator.codes.energyefficiency.ProjectEnergyEfficiencyContentCodes;
import com.persagy.apm.report.indicator.model.entity.EnergyEfficiencyDefine;
import com.persagy.apm.report.indicator.model.entity.EnergyEfficiencyQualifiedStandard;
import com.persagy.apm.report.indicator.utils.IndicatorUtils;
import com.persagy.apm.report.userconfig.function.model.Function;
import com.persagy.apm.report.userconfig.statisticsitems.enums.QualifiedEnum;
import com.persagy.apm.report.userconfig.statisticsitems.enums.StatisticCodeEnum;
import com.persagy.apm.report.userconfig.statisticsitems.model.StatisticItem;
import com.persagy.apm.report.userconfig.statisticsitems.service.IStatisticItemService;
import org.apache.commons.lang3.StringUtils;
import org.assertj.core.util.Lists;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import java.util.*;
import java.util.stream.Collectors;

/**
 * 能效情况工厂
 *
 * @author lixing
 * @version V1.0 2021/7/20 3:55 下午
 **/
@Service
public class EnergyEfficiencyFactory {
    @Autowired
    EnergyEfficiencyServiceImpl energyEfficiencyService;
    @Autowired
    IStatisticItemService statisticItemService;

    /**
     * 获取当月能效
     *
     * @param energyEfficiencyDefine 能效情况指标定义
     * @param projectId              项目id
     * @param reportMonth            报告月份
     * @return 当月能效
     * @author lixing
     * @version V1.0 2021/7/23 11:09 上午
     */
    public AttrValueVO getCurrentMonth(
            EnergyEfficiencyDefine energyEfficiencyDefine, String projectId, Date reportMonth) {
        String monthlyEnergyEfficiency = energyEfficiencyService.getMonthlyEnergyEfficiency(
                energyEfficiencyDefine.getTargetCode(), projectId, reportMonth);
        return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.currentMonth.name(),
                monthlyEnergyEfficiency, AttrValueTypeEnum.number_2);
    }

    /**
     * 获取合格标准
     *
     * @param function 指标对象
     * @return 合格标准
     * @author lixing
     * @version V1.0 2021/8/2 6:49 下午
     */
    public String getQualifiedStandardValue(Function function) {
        String qualifiedStandardStr = function.getQualifiedStandard();
        if (StringUtils.isBlank(qualifiedStandardStr)) {
            return null;
        }
        Gson gson = new Gson();
        EnergyEfficiencyQualifiedStandard energyEfficiencyQualifiedStandard = gson.fromJson(
                qualifiedStandardStr, EnergyEfficiencyQualifiedStandard.class);
        return energyEfficiencyQualifiedStandard.getStandardEfficiency();
    }

    /**
     * 获取合格标准
     *
     * @param function 指标对象
     * @return 合格标准
     * @author lixing
     * @version V1.0 2021/8/2 6:49 下午
     */
    public AttrValueVO getQualifiedStandard(Function function) {
        return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.qualifiedStandard.name(),
                getQualifiedStandardValue(function), AttrValueTypeEnum.str);
    }

    /**
     * 获取是否合格
     *
     * @param currentMonth  本月能效
     * @param standardValue 能效标准值
     * @return 是否合格
     * @author lixing
     * @version V1.0 2021/8/2 6:49 下午
     */
    public AttrValueVO getIsQualified(Double currentMonth, Double standardValue) {
        if (currentMonth == null || standardValue == null) {
            return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.isQualified.name(),
                    null, AttrValueTypeEnum.str);
        }
        // 获取能效标准判断条件
        StatisticItem statisticItem = statisticItemService.getItemByCode(StatisticCodeEnum.STANDARD_EFFICIENCY.name());
        Double qualifyResult = DataUtils.getQualifyResult(standardValue, statisticItem.getQualifyFormula(), currentMonth);

        QualifiedEnum qualifiedEnum = QualifiedEnum.getByType(qualifyResult);
        if (qualifiedEnum == null) {
            return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.isQualified.name(),
                    null, AttrValueTypeEnum.str);
        }
        return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.isQualified.name(),
                qualifiedEnum.getDesc(), AttrValueTypeEnum.str);
    }

    /**
     * 获取不合格率
     *
     * @param projectCount     总项目数
     * @param unqualifiedCount 不合格项目数
     * @return 不合格率
     * @author lixing
     * @version V1.0 2021/8/2 6:49 下午
     */
    public AttrValueVO getUnqualifiedProportion(Double projectCount, Double unqualifiedCount) {
        if (projectCount == null || unqualifiedCount == null) {
            return new AttrValueVO(AreaEnergyEfficiencyContentCodes.unqualifiedProportion.name(),
                    null, AttrValueTypeEnum.percent_2);
        }

        return new AttrValueVO(AreaEnergyEfficiencyContentCodes.unqualifiedProportion.name(),
                String.valueOf(DataUtils.doubleDivide(unqualifiedCount, projectCount)),
                AttrValueTypeEnum.percent_2, DataConstants.PERCENT);
    }

    /**
     * 获取上月能效
     *
     * @param energyEfficiencyDefine 能效情况指标定义
     * @param projectId              项目id
     * @param reportMonth            报告月份
     * @return 上月能效
     * @author lixing
     * @version V1.0 2021/7/23 11:09 上午
     */
    public AttrValueVO getLastMonth(
            EnergyEfficiencyDefine energyEfficiencyDefine, String projectId, Date reportMonth) {
        Date firstDayOfLastMonth = DateUtils.getFirstDayOfLastMonth(reportMonth);
        String monthlyEnergyEfficiency = energyEfficiencyService.getMonthlyEnergyEfficiency(
                energyEfficiencyDefine.getTargetCode(), projectId, firstDayOfLastMonth);
        return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.lastMonth.name(),
                monthlyEnergyEfficiency, AttrValueTypeEnum.number);
    }

    /**
     * 获取环比增幅
     *
     * @param chainGrowth 环比增幅
     * @param lastMonth   上月能效
     * @return 环比增幅
     * @author lixing
     * @version V1.0 2021/7/23 11:09 上午
     */
    public AttrValueVO getChainRange(
            Double chainGrowth, Double lastMonth) {
        return IndicatorUtils.getRange(ProjectEnergyEfficiencyContentCodes.chainRange.name(),
                chainGrowth, lastMonth, AttrValueTypeEnum.percent_2);
    }

    /**
     * 获取环比增量
     *
     * @param currentMonth 本月能效
     * @param lastMonth    上月能效
     * @return 环比增量
     * @author lixing
     * @version V1.0 2021/7/23 11:09 上午
     */
    public AttrValueVO getChainGrowth(
            Double currentMonth, Double lastMonth) {
        return IndicatorUtils.getGrowth(ProjectEnergyEfficiencyContentCodes.chainGrowth.name(),
                currentMonth, lastMonth, AttrValueTypeEnum.number_2);
    }

    /**
     * 获取月内每日能效
     *
     * @param dailyEnergyEfficiency 每日能效列表
     * @param reportMonth 报告月
     * @return 月内每日能效
     * @author lixing
     * @version V1.0 2021/7/23 11:09 上午
     */
    public AttrValueVO getDailyData(
            List<EnergyEfficiencyDataVO> dailyEnergyEfficiency, Date reportMonth) {
        /* 每日能效列表中不一定包含报告月所有时间的数据，如果不包含某天的数据，将这天的数据设置为空，添加到每日能效列表中 */
        // 获取当前能效列表中包含的日期
        if (dailyEnergyEfficiency == null) {
            dailyEnergyEfficiency = new ArrayList<>();
        }
        Set<String> timeSet = dailyEnergyEfficiency.stream().
                map(EnergyEfficiencyDataVO::getTime).collect(Collectors.toSet());

        // 遍历报告月日期，补全数据
        List<String> dayListOfMonth = DateUtils.getDayListOfMonth(reportMonth, DateUtils.SDF_SECOND);
        if (CollectionUtils.isEmpty(dayListOfMonth)) {
            throw new IllegalArgumentException("获取指定月份日期列表失败:" + reportMonth);
        }
        for (String day : dayListOfMonth) {
            if (!timeSet.contains(day)) {
                EnergyEfficiencyDataVO energyEfficiencyDataVO = new EnergyEfficiencyDataVO(day, null);
                dailyEnergyEfficiency.add(energyEfficiencyDataVO);
            }
        }

        Gson gson = new Gson();
        return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.dailyData.name(),
                gson.toJson(dailyEnergyEfficiency), AttrValueTypeEnum.list);
    }

    /**
     * 获取能效值最大日期列表
     *
     * @param dailyEnergyEfficiency 能效值列表
     * @return 能效值最大日期列表
     * @author lixing
     * @version V1.0 2021/7/29 10:13 上午
     */
    public AttrValueVO getMaxDays(List<EnergyEfficiencyDataVO> dailyEnergyEfficiency) {
        if (CollectionUtils.isEmpty(dailyEnergyEfficiency)) {
            return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.maxDays.name(), "", AttrValueTypeEnum.str);
        }
        // 将每日能效转换为能效值与日期列表的映射
        Map<Double, List<String>> efficiencyDateListMap = efficiencyDailyData2Map(dailyEnergyEfficiency);
        if (CollectionUtils.isEmpty(efficiencyDateListMap)) {
            return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.maxDays.name(), "", AttrValueTypeEnum.str);
        }
        // 将能效值降序排序
        List<Double> efficiencyList = Lists.newArrayList(efficiencyDateListMap.keySet());
        efficiencyList = efficiencyList.stream().sorted(Comparator.reverseOrder()).collect(Collectors.toList());

        Double maxValue = efficiencyList.get(0);
        List<String> maxDays = efficiencyDateListMap.get(maxValue);
        return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.maxDays.name(),
                String.join(",", maxDays), AttrValueTypeEnum.str);
    }

    /**
     * 获取能效值最小日期列表
     *
     * @param dailyEnergyEfficiency 能效值列表
     * @return 能效值最小日期列表
     * @author lixing
     * @version V1.0 2021/7/29 10:13 上午
     */
    public AttrValueVO getMinDays(List<EnergyEfficiencyDataVO> dailyEnergyEfficiency) {
        if (CollectionUtils.isEmpty(dailyEnergyEfficiency)) {
            return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.minDays.name(), "", AttrValueTypeEnum.str);
        }
        // 将每日能效转换为能效值与日期列表的映射
        Map<Double, List<String>> efficiencyDateListMap = efficiencyDailyData2Map(dailyEnergyEfficiency);
        if (CollectionUtils.isEmpty(efficiencyDateListMap)) {
            return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.minDays.name(), "", AttrValueTypeEnum.str);
        }
        // 将能效值升序排序
        List<Double> efficiencyList = Lists.newArrayList(efficiencyDateListMap.keySet());
        efficiencyList = efficiencyList.stream().sorted().collect(Collectors.toList());

        Double minValue = efficiencyList.get(0);
        List<String> minDays = efficiencyDateListMap.get(minValue);
        return new AttrValueVO(ProjectEnergyEfficiencyContentCodes.minDays.name(),
                String.join(",", minDays), AttrValueTypeEnum.str);
    }

    /**
     * 将每日能效转换为能效值与日期列表的映射
     *
     * @param dailyEnergyEfficiency 每日能效列表
     * @return {能效值 -> [日期]} 日期格式为 xx月xx日
     * @author lixing
     * @version V1.0 2021/7/29 10:26 上午
     */
    private Map<Double, List<String>> efficiencyDailyData2Map(
            List<EnergyEfficiencyDataVO> dailyEnergyEfficiency) {
        if (CollectionUtils.isEmpty(dailyEnergyEfficiency)) {
            return null;
        }
        Map<Double, List<String>> results = new HashMap<>();
        for (EnergyEfficiencyDataVO energyEfficiencyDataVO : dailyEnergyEfficiency) {
            String data = energyEfficiencyDataVO.getData() == null ?
                    null : String.valueOf(energyEfficiencyDataVO.getData());
            if (StringUtils.isBlank(data)) {
                continue;
            }
            List<String> tmp = results.computeIfAbsent(DataUtils.parseDouble(data), k -> new ArrayList<>());
            // 日期转换为xx月xx日格式
            String dateStr = null;
            try {
                dateStr = DateUtils.date2Str(
                        DateUtils.str2Date(energyEfficiencyDataVO.getTime(), DateUtils.SDF_SECOND),
                        DateUtils.SDF_CHAR_MONTH_DAY);
            } catch (Exception e) {
                throw new IllegalArgumentException("每日能效日期转换异常， 每日能效列表：" + dailyEnergyEfficiency);
            }
            tmp.add(dateStr);
        }
        return results;
    }

    /**
     * 获取预算信息
     *
     * @param function               指标对象
     * @param energyEfficiencyDefine 能效情况指标定义
     * @param projectId              项目id
     * @param reportMonth            报告月份
     * @return 预算信息
     * @author lixing
     * @version V1.0 2021/7/22 6:39 下午
     */
    public ItemVO getEnergyEfficiencyInfo(
            Function function, EnergyEfficiencyDefine energyEfficiencyDefine,
            String projectId, Date reportMonth) {
        List<AttrValueVO> attrValues = new ArrayList<>();
        ItemVO itemVO = new ItemVO(function.getId(), function.getName(), attrValues);

        List<EnergyEfficiencyDataVO> dailyEnergyEfficiency = energyEfficiencyService.getDailyEnergyEfficiency(
                energyEfficiencyDefine.getTargetCode(), projectId, reportMonth);
        // 本月能效值
        AttrValueVO currentMonth = getCurrentMonth(energyEfficiencyDefine, projectId, reportMonth);
        attrValues.add(currentMonth);
        // 上月能效值
        AttrValueVO lastMonth = getLastMonth(energyEfficiencyDefine, projectId, reportMonth);
        attrValues.add(lastMonth);
        // 环比增量
        AttrValueVO chainGrowth = getChainGrowth(
                DataUtils.parseDouble(currentMonth.getValue()), DataUtils.parseDouble(lastMonth.getValue()));
        attrValues.add(chainGrowth);
        // 环比增幅
        AttrValueVO chainRange = getChainRange(
                DataUtils.parseDouble(chainGrowth.getValue()), DataUtils.parseDouble(lastMonth.getValue()));
        attrValues.add(chainRange);
        // 能效最高日
        AttrValueVO maxDays = getMaxDays(dailyEnergyEfficiency);
        attrValues.add(maxDays);
        // 能效最低日
        AttrValueVO minDays = getMinDays(dailyEnergyEfficiency);
        attrValues.add(minDays);
        // 每日能效
        AttrValueVO dailyData = getDailyData(dailyEnergyEfficiency, reportMonth);
        attrValues.add(dailyData);

        // 合格标准
        AttrValueVO qualifiedStandard = getQualifiedStandard(function);
        attrValues.add(qualifiedStandard);

        // 是否合格
        AttrValueVO isQualified = getIsQualified(
                DataUtils.parseDouble(currentMonth.getValue()), DataUtils.parseDouble(qualifiedStandard.getValue()));
        attrValues.add(isQualified);

        return itemVO;
    }
}
